/* -*- Mode: C -*- */

// ===========================================================================
//
//   P O W E R    S U P P L Y    C O N T R O L L E R
//
//   This file:   C O N T R O L L E R . C
//
//   $Id: controller.c,v 1.6 2009/01/06 23:12:47 frgo Exp frgo $
//
// ===========================================================================

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>

#include <avr/wdt.h>
#include <avr/interrupt.h>
#include <avr/io.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <avr/pgmspace.h>

#include "gaul.h"

#include "controller.h"

// --------------------------------------------------------------------------
// GLOBAL DEFINITIONS
// --------------------------------------------------------------------------

// DEBUG "SWITCHES"

// #define __DEBUG__
// #define __DEBUG_IO__
// #define __DEBUG_DEBOUNCE__
// #define __DEBUG_WAIT_FOR_PBS__
// #define __DEBUG_BLINK_LAMPS__
// #define __DEBUG_ACCEPTED_PBS__
// #define __DEBUG_STANDBY__
// #define __DEBUG_OPERATE__
// #define __DEBUG_TEST_MODE__
// #define __DEBUG_GET_RELAY__
// #define __DEBUG_GET_VOLTAGE__
// #define __DEBUG_RELAY_ID_TO_I2C_ADDRESS__
// #define __DEBUG_SWITCH_RELAY__
// #define __DEBUG_GET_STATUS_TEXT__

// --------------------------------------------------------------------------
// GLOBAL VARS
// --------------------------------------------------------------------------

static char acPSUModel[ PSU_BUFFER_LENGTH + 1 ];
static char acPSUSerialNr[ PSU_BUFFER_LENGTH + 1 ];
static char acPSUVersion[ PSU_BUFFER_LENGTH + 1 ];

volatile uint8_t   nLock;
volatile uint16_t  nTimerLoadValue;

// PUSHBUTTON DEBOUNCING

volatile uint16_t key_state;  // debounced and inverted key state:
                              // bit = 1: key pressed
volatile uint16_t key_press;  // key press detect

volatile uint16_t nAcceptedPBs;
volatile uint16_t nPBSample; 

volatile bool_t bUseI2C = FALSE;

volatile uint16_t nCmdFromISR        = NO_CMD;
volatile bool_t   bInStatusNetEngine = FALSE;

// Init reason texts (each max 20 chars)

char acReasonPOR[]      = "-> PWR ON !";
char acReasonEXTRST[]   = "-> EXT RST !";
char acReasonBROWNOUT[] = "-> BROWN OUT !";
char acReasonWATCHDOG[] = "-> WATCHDOG !";

// STATUS NET

static tsStatusDefinition asStatuses[]  =
  { { PSU_STATUS_ID_PSU_OFF,           PSU_STATUS_TEXT_PSU_OFF           },
    { PSU_STATUS_ID_PSU_BOOTING,       PSU_STATUS_TEXT_PSU_BOOTING       },
    { PSU_STATUS_ID_PSU_STANDBY,       PSU_STATUS_TEXT_PSU_STANDBY       },
    { PSU_STATUS_ID_PSU_OPERATING,     PSU_STATUS_TEXT_PSU_OPERATING     },
    { PSU_STATUS_ID_PSU_IN_TEST_MODE,  PSU_STATUS_TEXT_PSU_IN_TEST_MODE  },
    { PSU_STATUS_ID_PSU_FAILURE,       PSU_STATUS_TEXT_PSU_FAILURE       },
    { PSU_STATUS_ID_PSU_SHUTTING_DOWN, PSU_STATUS_TEXT_PSU_SHUTTING_DOWN },
    { PSU_STATUS_ID_PSU_POWER_ON,      PSU_STATUS_TEXT_PSU_POWER_ON      } };

volatile int       nNrStatuses;
volatile tStatusID nCurrentStatus;
volatile tStatusID nPreviousStatus;
volatile tStatusID nNextStatus;

volatile uint16_t  nFailureReasonCode; // Read by fail()

volatile uint16_t  nDebugLevel = PSU_DEBUG_LEVEL_NONE; // Used by DEBUG_SHOW_STEP
// I2C CARDS

static tsI2CCard asI2CCards[]  =
  { { I2C_CARD_0, I2C_CARD_0_OUTPUT_ADDRESS, I2C_CARD_0_INPUT_ADDRESS } };

volatile int nNrI2CCards;

// RELAYS

static tsRelay asRelays[]  =
  { { RELAY_ID_UA,             I2C_CARD_0, RELAY_OFF, LAMP_ID_UA         },
    { RELAY_ID_UG2,            I2C_CARD_0, RELAY_OFF, LAMP_ID_UG2        },
    { RELAY_ID_UG1,            I2C_CARD_0, RELAY_OFF, LAMP_ID_UG1        },
    { RELAY_ID_UH,             I2C_CARD_0, RELAY_OFF, LAMP_ID_UH         },
    { RELAY_ID_HV_ENABLE,      I2C_CARD_0, RELAY_OFF, LAMP_ID_HV_ENABLE  },
    { RELAY_ID_MAIN_POWER,     I2C_CARD_0, RELAY_OFF, LAMP_ID_MAIN_POWER }, 
    { RELAY_ID_MAIN_FAN_SPEED, I2C_CARD_0, RELAY_OFF, NO_LAMP            } 
  };

volatile int nNrRelays;

// FANS

static tsFan asFans[]  =
  { { FAN_ID_MAIN_FAN, I2C_CARD_0, FAN_OFF } };

volatile int nNrFans;

// VOLTAGES

static tsVoltage asVoltages[]  =
  { { VOLTAGE_ID_UA,   VOLTAGE_ADC_MUX_0, VOLTAGE_OFF, 
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE },
    
    { VOLTAGE_ID_UG2,  VOLTAGE_ADC_MUX_1, VOLTAGE_OFF, 
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE },
    
    { VOLTAGE_ID_UG1,  VOLTAGE_ADC_MUX_2, VOLTAGE_OFF, 
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE },
    
    { VOLTAGE_ID_UH,   VOLTAGE_ADC_MUX_3, VOLTAGE_OFF, 
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE,
      VOLTAGE_RANGE_NOT_INITIALIZED_VALUE } };

volatile int nNrVoltages;

static uint8_t anVoltageCombinations[]  =
  { VOLTAGE_ID_UA | VOLTAGE_ID_UG2 | VOLTAGE_ID_UG1 | VOLTAGE_ID_UH ,
    VOLTAGE_ID_UH ,
    VOLTAGE_ID_UH | VOLTAGE_ID_UA | VOLTAGE_ID_UG1 };

volatile int nNrVoltageCombinations;
volatile bool_t bCheckVoltageCombination = TRUE;

// LAMPS

static tsLamp asLamps[]  =
  { { LAMP_ID_MAIN_POWER, I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_OPER,       I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_TEST,       I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_UA,         I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_UG2,        I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_UG1,        I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_UH,         I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_HV_ENABLE,  I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_RESET,      I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK },
    { LAMP_ID_FAIL,       I2C_CARD_0, LAMP_STATE_OFF, LAMP_DONT_BLINK } };

volatile int nNrLamps;

// PUSHBUTTONS

static tsPB asPBs[]  =
  { { PSU_FP_SW_MAIN_POWER, NO_I2C,     LAMP_ID_MAIN_POWER, RELAY_ID_MAIN_POWER },
    { PSU_FP_PB_OPER,       I2C_CARD_0, LAMP_ID_OPER,       NO_RELAY       },
    { PSU_FP_PB_TEST,       I2C_CARD_0, LAMP_ID_TEST,       NO_RELAY       },
    { PSU_FP_PB_UA,         I2C_CARD_0, LAMP_ID_UA,         VOLTAGE_ID_UA  },
    { PSU_FP_PB_UG2,        I2C_CARD_0, LAMP_ID_UG2,        VOLTAGE_ID_UG2 },
    { PSU_FP_PB_UG1,        I2C_CARD_0, LAMP_ID_UG1,        VOLTAGE_ID_UG1 },
    { PSU_FP_PB_UH,         I2C_CARD_0, LAMP_ID_UH,         VOLTAGE_ID_UH  },
    { PSU_FP_PB_HV_ENABLE,  I2C_CARD_0, LAMP_ID_HV_ENABLE,  RELAY_ID_HV_ENABLE },
    { PSU_FP_PB_RESET,      I2C_CARD_0, LAMP_ID_RESET,      NO_RELAY } };

volatile int nNrPBs;

// --------------------------------------------------------------------------
// GENERAL UTILITY FUNCTIONS
// --------------------------------------------------------------------------

uint16_t lock ( void )
{
  uint8_t  nTmpSREG = SREG; // Save status register
  uint16_t nRC      = RC_OK;
  
  cli();                   // Ensure atomic execution
  
  if( nLock == 0 )
  {
    nLock = 1;
  }
  else 
  {
    nRC = RC_ALREADY_LOCKED;
  }
    
  SREG = nTmpSREG;         // We're done so interrupts ok now
  
  return nRC;  
}

void unlock ( void )
{
  uint8_t  nTmpSREG = SREG; // Save status register
  
  cli();                   // Ensure atomic execution
  
  if( nLock == 1 )
  {
    nLock = 0;
  }
  
  SREG = nTmpSREG;         // We're done so interrupts ok now
  
  return;  
}

// --------------------------------------------------------------------------
// FAILURE MANAGEMENT
// --------------------------------------------------------------------------

rc_t set_failure_reason_code( uint16_t nReasonCode )
{
  nFailureReasonCode = nReasonCode;

  return RC_OK;
}

uint16_t get_failure_reasoncode( void )
{
  return nFailureReasonCode;
}

rc_t print_failure_reasoncode( FILE * stream )
{
  rc_t nRC = RC_OK;

  tsDARCEvent sEvent;
  
  sEvent.nType              = DARC_EVENT_TYPE_FAILURE_MESSAGE;
  sEvent.nFailureReasonCode = nFailureReasonCode;

  nRC = darc_send_event( &sEvent );
   
  return nRC;
}

// --------------------------------------------------------------------------
// GENERAL I/O FUNCTIONS
// --------------------------------------------------------------------------

void init_io_timer ( void )
{
  uint8_t  nTmpSREG = SREG; // Save status register
 
  // We use timer 1 (16 bit Timer) in CTC mode
  // This timer shall generate an interrupt every PB_READ_INTERVAL seconds
  // Prescaler has to be set to 1 !

  double   nTickInterval         = 0.0; // [s]
  double   nDesiredTimerInterval = PSU_FP_PB_READ_INTERVAL;
  uint16_t nPrescalerValue       = PSU_IO_TIMER_PRESCALER_VALUE;

  cli();                   // Ensure atomic execution
  
  nTickInterval   = 1 / F_CPU * nPrescalerValue;
  nTimerLoadValue = nDesiredTimerInterval / nTickInterval - 1; 
    
  OCR1A  = nTimerLoadValue; // Load the timer compare value
  
  TCNT1  = 0;               // Reset counter value
  TCCR1A = 0;               // Keine PWM
  TIMSK  |= _BV( OCIE1A );  // Output Compare A Interrupt: enable it
  TCCR1B |= _BV( WGM12 );   // Set CTC mode
  
  TCCR1B &= ~( _BV( CS12 )); // Next three lines:
  TCCR1B &= ~( _BV( CS11 )); // Set prescaler to 1
  TCCR1B |= _BV( CS10 );     // and start timer;

  asm ( "nop" "\r\n"
        "nop" "\r\n"
        "nop" "\r\n"
        "nop" "\r\n"
        "nop" "\r\n" );

  SREG = nTmpSREG;         // We're done so interrupts ok now

  return;
}

// --------------------------------------------------------------------------
// DARC I/O
// --------------------------------------------------------------------------

// DARC convenience functions to send various "object's" info ...

rc_t darc_send_psu_status( uint8_t nStatus )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    tsDARCEvent sEvent;
 
    sEvent.nType        = DARC_EVENT_TYPE_STATUS_CHANGE; 
    sEvent.pcStatusText = get_status_text( nStatus );

    nRC = darc_send_event( &sEvent );
  }

  return nRC;
}

rc_t darc_send_errorlog( void )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_ERRORLOG, 0, darc_stream() );
  }

  return nRC;
}

rc_t darc_send_relay( uint8_t nID )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_RELAY, nID, darc_stream() );
  }

  return nRC;
}

rc_t darc_send_voltage( uint8_t nID )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_VOLTAGE, nID, darc_stream() );
  }

  return nRC;
}

rc_t darc_send_fan( uint8_t nID )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_FAN, nID, darc_stream() );
  }

  return nRC;
}

rc_t darc_send_lamp( uint8_t nID )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_LAMP, nID, darc_stream() );
  }

  return nRC;
}

rc_t darc_send_failure_reasoncode( void )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_FAILURE_REASONCODE, 0, darc_stream() );
  }

  return nRC;
}

rc_t darc_send_psu_configuration( void )
{
  rc_t nRC = RC_OK;

  if( use_darc() == TRUE )
  {
    nRC = print_object( OBJECT_CLASS_ID_PSU_CONFIGURATION, 0, darc_stream() );    
  }

  return nRC;
}

rc_t print_psu_configuration( FILE * stream )
{
  rc_t nRC = RC_OK;

  if(!ferror( stream ))
  {
    fprintf( stream, "PSU CONFIG:\n" );
  
    nRC = print_relays( stream );
    
    if( ! IS_ERROR( nRC ) )
      nRC = print_voltages( stream );
    
    if( ! IS_ERROR( nRC ) )
      nRC = print_fans( stream );
    
    if( ! IS_ERROR( nRC ) )
      nRC = print_lamps( stream );
    
    fprintf( stream, "\n%c\n", DARC_CAT_CMD_END_CHAR );
  }
  else
  {
    nRC = RC_ERROR_STREAM_ERROR;
    log_error( nRC, __LINE__ );
  }

  return nRC;
}

rc_t darc_receive_command( tpsDARCCommandPtr psCmd )
{
  rc_t     nRC       = RC_OK;
  char    *pcCmd;
  int      nCmdLen;

  char     acToken[ DARC_CAT_CMD_MAX_LENGTH + 1 ];
  int      nTokenLen;

  char    *pcPos;
  char    *pcPos2;

  bool_t   bCmdDone  = FALSE;

  nRC = darc_receive();

  if( nRC == RC_BUSY )
    return DARC_NO_CMD;

  // Parse command

  pcCmd = darc_receive_buffer();

  nCmdLen = strlen( pcCmd );
  
  if( nCmdLen <= strlen( DARC_CAT_CMD_PREFIX ) )
  {
    // => Invalid command.
    darc_clear_receive_buffer();
    darc_send_result( RC_ERROR );

    return DARC_NO_CMD;
  }

  strupr( pcCmd );  // convert all to upper case letter

  if( ! strncmp( pcCmd, DARC_CAT_CMD_PREFIX, strlen( DARC_CAT_CMD_PREFIX )))
  {
    // => Not a DARC command
    darc_clear_receive_buffer();

    return DARC_NO_CMD;
  }
  
  pcPos = strchr( pcCmd, ' '); // Positions after first space
  pcPos += sizeof( char );

  memset( acToken, '\0', DARC_CAT_CMD_MAX_LENGTH + 1 );

  pcPos2 = strchr( pcPos, ' '); // Next space is end of command

  nTokenLen = (pcPos2 - pcPos ) / sizeof( char );

  // Cmd #1/#2 : SWITCH-VOLTAGE

  // Cmd #3/#4 : SET / GET VOLTAGE 

  // Cmd #5 : PSU-STATE

  // Cmd #6 : ENABLE-HV

  // Cmd #7 : DISABLE-HV

  if( ! strncmp( pcPos, DARC_CAT_CMD_HV_ENABLED, nTokenLen) )
  {
    bCmdDone = TRUE;
    psCmd->nID = DARC_CAT_CMD_ID_GET_HV_STATE;
    goto EXIT_DARC_RECEIVE_COMMAND;
  }

  // Cmd #8 : HV-ENABLED?

  if( ! strncmp( pcPos, DARC_CAT_CMD_HV_ENABLED, nTokenLen) )
  {
    bCmdDone = TRUE;
    psCmd->nID = DARC_CAT_CMD_ID_GET_HV_STATE;
    goto EXIT_DARC_RECEIVE_COMMAND;
  } 

  // Cmd #9 : PSU-RESET

  if( ! strncmp( pcPos, DARC_CAT_CMD_PSU_RESET, nTokenLen) )
  {
    bCmdDone = TRUE;
    psCmd->nID = DARC_CAT_CMD_ID_PSU_RESET;
    goto EXIT_DARC_RECEIVE_COMMAND;
  } 

  // Cmd #10 : PSU-SHUTDOWN

  if( ! strncmp( pcPos, DARC_CAT_CMD_PSU_SHUTDOWN, nTokenLen) )
  {
    bCmdDone = TRUE;
    psCmd->nID = DARC_CAT_CMD_ID_PSU_SHUTDOWN;
    goto EXIT_DARC_RECEIVE_COMMAND;
  } 

 EXIT_DARC_RECEIVE_COMMAND:

  darc_clear_receive_buffer();
  
  if( ! bCmdDone )
  { 
    // => Invalid command.
    darc_send_result( RC_ERROR );

    return DARC_NO_CMD;
  }

  return nRC;
}

rc_t process_darc( void )
{
  rc_t nRC = RC_OK;

  tsDARCCommand sCmd;

  nRC = darc_receive_command( &sCmd );
   
  if( nRC != DARC_NO_CMD )
  {
    if( ! IS_ERROR( nRC ) )
    {
      nRC = process_darc_command( &sCmd );
    }
    else
      nRC = RC_OK;
  }
  else
    nRC = RC_OK;
 
  return nRC;
}

rc_t process_darc_command( tpsDARCCommandPtr psCmd )
{
  rc_t nRC = RC_OK;

  switch( psCmd->nID )
  {
    case DARC_CAT_CMD_ID_SWITCH_VOLTAGE_OFF:

#if 0            
      if( get_status() == PSU_STATUS_ID_PSU_IN_TEST_MODE )
        nRC = darc_send_result( switch_voltage( psCmd, VOLTAGE_OFF ) );    
      else
#endif
        nRC = darc_send_result( RC_ERROR_OPERATION_NOT_PERMITTED );
      break;
                                
    case DARC_CAT_CMD_ID_SWITCH_VOLTAGE_ON:

#if 0                                
      if( get_status() == PSU_STATUS_ID_PSU_IN_TEST_MODE )
        nRC = darc_send_result( switch_voltage( , VOLTAGE_OFF ) );    
      else
#endif
        nRC = darc_send_result( RC_ERROR_OPERATION_NOT_PERMITTED );

      break;
                               
    case DARC_CAT_CMD_ID_GET_VOLTAGE:
                                
      nRC = darc_send_result( RC_ERROR_FUNCTION_NOT_IMPLEMENTED );
      break;
                                
    case DARC_CAT_CMD_ID_SET_VOLTAGE:
                            
      nRC = darc_send_result( RC_ERROR_FUNCTION_NOT_IMPLEMENTED );
      break;  
                                
    case DARC_CAT_CMD_ID_PSU_GET_STATE:
                                
      nRC = darc_send_result( darc_send_psu_status( get_status()));
      break;
                              
    case DARC_CAT_CMD_ID_ENABLE_HV:
                                
      nRC = darc_send_result( RC_ERROR_FUNCTION_NOT_IMPLEMENTED );
      break;
                                
    case DARC_CAT_CMD_ID_DISABLE_HV:
                            
      nRC = darc_send_result( RC_ERROR_FUNCTION_NOT_IMPLEMENTED );
      break;

    case DARC_CAT_CMD_ID_GET_HV_STATE:
                               
      nRC = darc_send_result( RC_ERROR_FUNCTION_NOT_IMPLEMENTED );
      break;
                                
    case DARC_CAT_CMD_ID_PSU_RESET:            
                                
      nRC = darc_send_result( switch_status( PSU_STATUS_ID_PSU_RESETTING ) );
      break;

    case DARC_CAT_CMD_ID_PSU_SHUTDOWN:            
                           
      nRC = darc_send_result( switch_status( PSU_STATUS_ID_PSU_SHUTTING_DOWN ) );
      break;

    default:

      nRC = darc_send_result( RC_ERROR_INVALID_DARC_COMMAND );
      break;                            
  }

  return nRC;
}

// --------------------------------------------------------------------------
// STATUS NET MANAGEMENT
// --------------------------------------------------------------------------

tStatusID get_status( void )
{
  return nCurrentStatus;
}

char *get_status_text( tStatusID nStatus )
{
  uint8_t  nI;
  char    *pcText = NULL;

  if( nStatus == PSU_STATUS_ID_NO_STATUS )
    return NULL;  

  for( nI = 0; nI < nNrStatuses; nI++ )
  {
    if( asStatuses[ nI ].nID == nStatus )
    {
      pcText = asStatuses[ nI ].pcText;
      break;
    }
  }

  return pcText;
}

void update_status_display( void )
{
  tStatusID  nStatus;
  char      *pcStatusText;

  uint16_t  nErrorNr;
  uint16_t  nErrorLine;

  nStatus      = get_status();
  pcStatusText = get_status_text( nStatus );

  // Update status on LCD

  if ( pcStatusText != (char *) NULL )
  {
    nErrorNr   = get_error_nr();
    nErrorLine = get_error_line();

    display_format( DISPLAY_STATUS_LINE,
                    "* %-6s * %04d-%04d", 
                    pcStatusText, 
                    (int) nErrorNr, 
                    (int) nErrorLine );
  }
 
  darc_send_psu_status( nStatus );

  return;
}

void set_status( tStatusID nStatus )
{
  uint8_t nTmpSREG = SREG;

  cli();

  nPreviousStatus = nCurrentStatus;
  nCurrentStatus  = nStatus;

  update_status_display();

  SREG = nTmpSREG;

  return;
}

rc_t switch_status( tStatusID nNext )
{
  nNextStatus = nNext;

  return RC_OK;
}

rc_t goto_next_status( tStatusID nNext )
{
  rc_t nRC = RC_OK;
  
  switch( nNext )
  {
    case PSU_STATUS_ID_PSU_OFF:

      nRC = psu_off();

      break;

    case PSU_STATUS_ID_PSU_BOOTING:

      nRC = boot();

      break;

    case PSU_STATUS_ID_PSU_STANDBY:

      nRC = standby();            

      break;

    case PSU_STATUS_ID_PSU_OPERATING:

      nRC = operate();            

      break;

    case PSU_STATUS_ID_PSU_IN_TEST_MODE:

      nRC = test_mode();            

      break;

    case PSU_STATUS_ID_PSU_FAILURE:
   
      nRC = fail( get_failure_reasoncode() );

      break;

    case PSU_STATUS_ID_PSU_SHUTTING_DOWN:
   
      nRC = shutdown();

      break;

    case PSU_STATUS_ID_PSU_RESETTING:
   
      reset();
      // This point never reached ...
      break;

    default:

      nRC = RC_ERROR_ILLEGAL_STATUS_ID;
      break;  
  }

  return nRC;
}

void heart_beat( void )
{
  static uint32_t nTicks;
    
  nTicks++;

  if( nTicks > HEART_BEAT_TICKS )
  {
    if( bit_is_set( PINB, PINB0 ) )
    {
      PORTB = PINB & ~( 1 << PINB0 );
    } 
    else
    {
      PORTB = PINB | (1 << PB0);
    }
    nTicks = 0;
  }
  return;
}

rc_t print_status( FILE * stream )
{
  rc_t nRC = RC_OK;

  char *pcStatusText = get_status_text( get_status() );

  if(( stream != NULL ) && (!ferror( stream )))
  {
    fprintf( stream, "STATUS %s.\n", pcStatusText );
  }
  else
  {
    nRC = RC_ERROR_STREAM_ERROR;
    log_error( nRC, __LINE__ );
  }

  return nRC;
}

// --------------------------------------------------------------------------
// PUSH BUTTON CONTROL
// --------------------------------------------------------------------------

uint16_t get_pb_sample( void )
{ 
  uint16_t nResult;
  uint8_t  nPortSample;
  
  uint8_t  nTmpSREG    = SREG; // Save status register

  cli();                       // Ensure atomic execution

  nPortSample = PINA;          // Read the RESET PB

  nResult = (nPortSample << 8);  // Save in higher 8 bits. => RESET = Bit 8 

  if( bUseI2C == TRUE )
  {
    nPortSample = i2c_read( I2C_ADDRESS_FP_PB ); // Read all other PBs 
                                                 // (Bits 0 to 7)
  }
  
  nResult += nPortSample;      // And save in lower 8 bits

  SREG = nTmpSREG;             // We're done so interrupts ok now 

  return nResult;
}

void debounce( void )      // Needs to be called every 10ms !
{ 
  static uint16_t ct0 = 0;
  static uint16_t ct1 = 0;

  uint16_t i = 0;  
  uint16_t nSample = get_pb_sample();
  
  uint8_t  nTmpSREG = SREG;
  
  cli();

  i = key_state ^ ~nSample;                       // key changed ?
  ct0 = ~( ct0 & i );                             // reset or count ct0
  ct1 = ct0 ^ (ct1 & i);                          // reset or count ct1
  i &= ct0 & ct1;                                 // count until roll over ?
  key_state ^= i;                                 // then toggle debounced state
  key_press |= key_state & i;                     // 0->1: key press detect

#if defined( __DEBUG_DEBOUNCE__ )
  display_format( 1, "s = %d, p = %d", key_state, key_press );
  display_format( 2, "Sample = %d", nSample );
#endif
  
  SREG = nTmpSREG;
  
  return; 
}

rc_t set_accepted_pbs( uint16_t nPBMask )
{
  rc_t nRC = RC_OK;

  nAcceptedPBs = nPBMask;

#if defined( __DEBUG_ACCEPTED_PBS__ )
  display_format( 4, "AcceptedPBs = %u", nAcceptedPBs );
#endif

  return nRC;
}

uint16_t get_pbs( uint16_t nPBMask )
{
  uint8_t nTmpSREG = SREG;
  cli();                                          // read and clear atomic !

  nPBMask   &= key_press;                          // read key(s)
  key_press ^= nPBMask;                          // clear key(s)

  SREG = nTmpSREG;

  return key_state;
}

void clear_key_state( void )
{
  uint8_t nTmpSREG = SREG;
  
  cli();

  key_state = 0;
  key_press = 0;

  SREG = nTmpSREG;
}

uint16_t wait_for_pbs( uint16_t nPBMask )
{
  bool_t   bFound   = FALSE;
  uint16_t nPBs;
  uint16_t nPBState;

  while( bFound == FALSE )
  {
    nPBState = get_pbs( nPBMask );
    nPBs = ~nPBState;

#if defined( __DEBUG_WAIT_FOR_PBS__ )
    display_format( 1, "nPBState = %u", nPBState );
    display_format( 2, "nPBMask = %u", nPBMask );
    display_format( 3, "nPBs = %u", nPBs ); 
    _delay_ms( 200 );
#endif

    if( nPBs & nPBMask )
    {
#if defined( __DEBUG_WAIT_FOR_PBS__ )
      display_format( 3, "nPBs = %u FOUND!", nPBs );
      _delay_ms( 200 );
#endif
      bFound = TRUE;
    }
    else
      _delay_ms( 1 );
  }
   
  return nPBs;
}

uint16_t wait_for_valid_pb( void )
{
  return wait_for_pbs( nAcceptedPBs );
}

void wait_for_power_on( void )
{
  uint16_t nPowerOn = 0;

  while( nPowerOn == 0 )
  {
    asm volatile ( "nop" );
     
    nPowerOn = ~key_state & PSU_FP_SW_MAIN_POWER;

    if( nPowerOn == 0 )
      _delay_ms( 1 );
  } 
 
  // Power switch on -> Switch on main power lamp and main power relay

  switch_relay( RELAY_ID_MAIN_POWER, RELAY_ON );
}

void wait_for_power_off( void )
{
  uint16_t nPowerStillOn = 1;  

  while( nPowerStillOn == 1 )
  {
    asm volatile ( "nop" );

    nPowerStillOn = ~key_state & PSU_FP_SW_MAIN_POWER;

    if( nPowerStillOn )
      _delay_ms( 1 );  
  } 

  switch_relay( RELAY_ID_MAIN_POWER, RELAY_OFF );
 
  return;    
}
// --------------------------------------------------------------------------
// I2C CARD UTILS
// --------------------------------------------------------------------------

tpsI2CCardPtr get_i2ccard( uint16_t nI2CCardID )
{
  tpsI2CCardPtr psI2CCard = NULL;

  for( uint16_t nI = 0; nI < nNrI2CCards; nI++ )
  {
    if( asI2CCards[ nI ].nID == nI2CCardID )
    {
      psI2CCard = &asI2CCards[ nI ];
      break;
    }
  }
  
  return psI2CCard;
}

rc_t switch_io( uint8_t nI2CAddress, uint8_t nBit, uint8_t nState )
{
  rc_t     nRC    = RC_OK;
  uint8_t  nValue;

  if( bUseI2C == TRUE )
  {
    if( IS_VALID_I2C_ADDRESS( nI2CAddress ) )
    {
      nValue = i2c_read_byte( nI2CAddress );

      switch( nState )
      {
        case IO_OFF:
        
          BIT_OFF(nValue,nBit);
          break;
        
        case IO_ON:

          BIT_ON(nValue,nBit);
          break;

        default:
          nRC = RC_ERROR_INVALID_RELAY_STATE;
          break;
      }

      if ( nRC == RC_OK )
      {
        i2c_write_byte( nI2CAddress, nValue ); 
      }
    }
  }
  
  return nRC;
}

// --------------------------------------------------------------------------
// FAN CONTROL
// --------------------------------------------------------------------------

tpsFanPtr get_fan( uint8_t nFanID )
{
  tpsFanPtr psFan = NULL;

  for( uint16_t nI = 0; nI < nNrFans; nI++ )
  {
    if( asFans[ nI ].nID == nFanID )
    {
      psFan = &asFans[ nI ];
      break;
    }
  }

  return psFan;
}

rc_t switch_fan( uint8_t nFanID, uint8_t nFanSpeed )
{
  rc_t nRC = RC_OK;

  tpsFanPtr psFan = get_fan( nFanID );

  if( psFan != NULL )
  {
    switch( nFanSpeed )
    {
      case FAN_OFF:
        // Do nothing - Fan always runs with FAN_SPEED_1
        break;
        
      case FAN_SPEED_1:
        nRC = switch_relay( psFan->nRelayID, RELAY_OFF );  
        break;

      case FAN_SPEED_2: // Same as FAN_SPEED_MAX !
        nRC = switch_relay( psFan->nRelayID, RELAY_ON );    
        break;
  
      default:
        nRC = RC_ERROR_INVALID_FAN_SPEED;
        break;
    }    
  }
  else
  {
    nRC = RC_ERROR_INVALID_IO_ID;
    log_error( nRC, __LINE__ );
  }

  return nRC;
}

rc_t print_fan( uint8_t nID, FILE *stream )
{
  rc_t nRC = RC_OK;

  tpsFanPtr psFan = get_fan( nID );

  if( psFan != NULL )
  {
    if(( stream != NULL ) && (!ferror( stream )))
    {
      fprintf( stream, "FAN ID=%d ", nID );
      fprintf( stream, "STATE=%d.\n ", psFan->nState );
    }
    else
    {
      nRC = RC_ERROR_STREAM_ERROR;
      log_error( nRC, __LINE__ );
    }
  }
  else
  {
    nRC = RC_ERROR_INVALID_IO_ID;
    log_error( nRC, __LINE__ );
  } 
  
  return nRC;
}

rc_t print_fans( FILE * stream )
{
  rc_t nRC = RC_OK;

  if(( stream != NULL ) && (!ferror( stream )))
  {
    fprintf( stream, "FANS AMOUNT=%d\n", nNrFans );

    for( uint8_t nI = 0; nI < nNrFans; nI++ )
    {
      nRC = print_fan( nI, stream );
      if( nRC != RC_OK )
        break;
    }

    fprintf( stream, "%c", DARC_CAT_CMD_END_CHAR );
  }
  else
  {
    nRC = RC_ERROR_STREAM_ERROR;
    log_error( nRC, __LINE__ );
  }

  return nRC; 
}

// --------------------------------------------------------------------------
// RELAY CONTROL
// --------------------------------------------------------------------------

tpsRelayPtr get_relay( uint16_t nRelayID )
{
  tpsRelayPtr psRelay = NULL;
  uint16_t    nI;

  if( nRelayID == NO_RELAY )
    return NULL;
 
  for( nI = 0; nI < nNrRelays; nI++ )
  {
    if( asRelays[ nI ].nID == nRelayID )
    {
      psRelay =  &asRelays[ nI ];
      break;
    }
  }

  return psRelay;
}

uint16_t relay_id_to_i2c_address( uint16_t nRelayID )
{
  tpsI2CCardPtr psI2CCard = NULL;
  uint16_t      nAddress  = 0;
  tpsRelayPtr   psRelay   = get_relay( nRelayID );

  if( psRelay != NULL )
  {
    psI2CCard = get_i2ccard( psRelay->nI2CCardID );
    
    if( psI2CCard != NULL )
      nAddress = psI2CCard->nOutAddress;
    else
      nAddress = RC_ERROR_INVALID_I2C_ADDRESS;
  }

#if defined( __DEBUG_RELAY_ID_TO_I2C_ADDRESS__ )
  display_format( 2, "Relay I2C Addr = %d", nAddress );
   _delay_ms( 2000 );
#endif

  return nAddress;
}

rc_t switch_relay( uint16_t nID, uint8_t nSwitchValue )
{
  rc_t        nRC     = RC_OK;
  tpsRelayPtr psRelay;

  if( nID == NO_RELAY )
    return RC_OK;

#if defined( __DEBUG_SWITCH_RELAY__ )
  display_format( 2, "Switch relay = %d", nID );
#endif

  psRelay = get_relay( nID );

  if( psRelay != NULL ) // Not NULL -> Relay found
  { 
    nRC = switch_io( relay_id_to_i2c_address( nID ), nID, nSwitchValue );
    if( nRC == RC_OK )
    {
      psRelay->nState = nSwitchValue;
      set_lamp_status( psRelay->nLampID, nSwitchValue ); 
    }
  }

  return nRC;
}

rc_t init_relays( void )
{
  rc_t    nRC = RC_OK;
  uint16_t nI;
  
  for( nI = 0; nI < nNrRelays; nI++ )
  {
    switch_relay( nI, RELAY_OFF );
  }
  
  return nRC;
}

rc_t toggle_relay( uint16_t nRelayID )
{
  rc_t        nRC     = RC_OK;
  tpsRelayPtr psRelay = get_relay( nRelayID );
 
  if( psRelay != NULL )
  {
    if( psRelay->nState == RELAY_OFF )
      nRC = switch_relay( nRelayID, RELAY_ON );
    else
      nRC = switch_relay( nRelayID, RELAY_OFF );
  }

  return nRC;  
}

rc_t print_relay( uint16_t nID, FILE *stream )
{
  rc_t nRC = RC_OK;

  tpsRelayPtr psRelay = get_relay( nID );

  if( psRelay != NULL )
  {
    if(( stream != NULL ) && (!ferror( stream )))
    {
      fprintf( stream, "<relay id=%d>", nID );
      fprintf( stream,   "<state>%s</state>", psRelay->nState == RELAY_ON ? "ON" : "OFF" );
      fprintf( stream, "</relay>" );
    }
    else
    {
      nRC = RC_ERROR_STREAM_ERROR;
      log_error( nRC, __LINE__ );
    }
  }
  
  return nRC;
}

rc_t print_relays( FILE * stream )
{
  rc_t nRC = RC_OK;

  if(( stream != NULL ) && (!ferror( stream )))
  {
    fprintf( stream, "<relays amount=%d>", nNrRelays );

    for( uint16_t nI = 0; nI < nNrRelays; nI++ )
    {
      nRC = print_relay( nI, stream );
      if( nRC != RC_OK )
        break;
    }

    fprintf( stderr, "</relays>" );
  }
  else
  {
    nRC = RC_ERROR_STREAM_ERROR;
    log_error( nRC, __LINE__ );
  }

  return nRC; 
}

// --------------------------------------------------------------------------
// VOLTAGE CONTROL
// --------------------------------------------------------------------------

tpsVoltagePtr get_voltage( uint16_t nID )
{
  tpsVoltagePtr psVoltage = NULL;
  uint16_t      nI;

  if( nID == NO_VOLTAGE )
    return NULL;

  for( nI = 0; nI < nNrVoltages; nI++ )
  {
    if( asVoltages[ nI ].nID == nID )
    {
      psVoltage =  &asVoltages[ nI ];
      break;
    }
  }

  return psVoltage;
}

rc_t switch_voltage( uint16_t nVoltageID, uint8_t nMode )
{
  rc_t          nRC       = RC_OK;
  tpsVoltagePtr psVoltage = get_voltage( nVoltageID );

  if( psVoltage != NULL )
  {

    nRC = switch_relay( nVoltageID, nMode );
    if( nRC == RC_OK )
    {
      psVoltage->nState = nMode;
    }
  }

  return nRC;
}

rc_t toggle_voltage( uint16_t nVoltageID )
{
  rc_t          nRC       = RC_OK;
  tpsVoltagePtr psVoltage = get_voltage( nVoltageID );
 
  if( psVoltage != NULL )
  {
    if( psVoltage->nState == VOLTAGE_OFF )
      nRC = switch_voltage( nVoltageID, VOLTAGE_ON );
    else
      nRC = switch_voltage( nVoltageID, VOLTAGE_OFF );
  }

  return nRC;  
}

rc_t switch_off_all_voltages( void )
{
  rc_t nRC = RC_OK;

  nRC = bring_down_voltages();

  return nRC;
}

rc_t bring_up_voltages( void )
{
  rc_t nRC = RC_OK;

  // Switch on voltages individually in the PSU.
  // ORDER OF EXECUTION MATTERS HERE !

  // Step 1: Heater Voltage
  nRC = switch_voltage( VOLTAGE_ID_UH,  VOLTAGE_ON );
  RESET_ON_ERROR(nRC);
  
  // Step 2: Anode Voltage 
  nRC = switch_voltage( VOLTAGE_ID_UA,  VOLTAGE_ON );
  RESET_ON_ERROR(nRC);

  // Step 3: Screen voltage  
  nRC = switch_voltage( VOLTAGE_ID_UG2, VOLTAGE_ON );
  RESET_ON_ERROR(nRC);

  // Step 4: Grid voltage
  nRC = switch_voltage( VOLTAGE_ID_UG1, VOLTAGE_ON ); 
  RESET_ON_ERROR(nRC);

  // Switch on the HV ENABLE line line. 

  nRC = switch_relay( RELAY_ID_HV_ENABLE, RELAY_ON );
  RESET_ON_ERROR(nRC);

  return nRC;
}

rc_t bring_down_voltages( void )
{
  rc_t nRC = RC_OK;

  // Switch off the HV ENABLE line first. This should kill
  // all voltages based on the controller board in the amp.

  nRC = switch_relay( RELAY_ID_HV_ENABLE, RELAY_OFF );
  RESET_ON_ERROR(nRC);

  // Then switch of voltages individually in the PSU
  // ORDER OF EXECUTION MATTERS HERE !
  
  // Step 1: Anode Voltage 
  nRC = switch_voltage( VOLTAGE_ID_UA,  VOLTAGE_OFF );
  RESET_ON_ERROR(nRC);

  // Step 2: Screen voltage  
  nRC = switch_voltage( VOLTAGE_ID_UG2, VOLTAGE_OFF );
  RESET_ON_ERROR(nRC);

  // Step 3: Grid voltage
  nRC = switch_voltage( VOLTAGE_ID_UG1, VOLTAGE_OFF ); 
  RESET_ON_ERROR(nRC);

  // Step 4: Heater Voltage
  nRC = switch_voltage( VOLTAGE_ID_UH,  VOLTAGE_OFF );
  RESET_ON_ERROR(nRC);

  return nRC;
}

void  set_check_voltage_combination_flag( bool_t bFlag )
{
  bCheckVoltageCombination = bFlag;
}

rc_t check_voltage_combination( uint16_t nVoltageID, uint8_t nMode )
{
  rc_t    nRC          = RC_OK;
  uint8_t nCombination = 0;
  bool_t  bValid       = FALSE;

  tpsVoltagePtr psVoltage;
  tpsVoltagePtr psNewVoltage = NULL;

  if( bCheckVoltageCombination != FALSE )
  {
    for( uint16_t nI = 0; nI < nNrVoltages; nI++ )
    { 
      psVoltage = &asVoltages[ nI ];
      nCombination |= (psVoltage->nState << psVoltage->nID);
      if( psVoltage->nID == nVoltageID )
      {
        psNewVoltage = psVoltage;
        break;
      }
    } 
    
    if( psNewVoltage != NULL )
    {
      if( psNewVoltage->nState != nMode ) // if a new state/mode 
      {
        if( nMode == VOLTAGE_ON )
          nCombination |= (VOLTAGE_ON << psNewVoltage->nID);
        else
          nCombination &= ~(VOLTAGE_OFF << psNewVoltage->nID); 

        // Now check if that new combination is listed as a valid one 

        for( uint8_t nI = 0; nI < nNrVoltageCombinations; nI++ )
        {
          if( nCombination == anVoltageCombinations[ nI ] )
          {
            bValid = TRUE;
            break;
          }
        }
        
        if( bValid  == FALSE )
          nRC = RC_ERROR_INVALID_VOLTAGE_COMBINATION;
      }
    }
    else
    {
      log_error( SW_ERROR, __LINE__ );
    }
  }

  return nRC;
}

rc_t set_voltage_range( uint16_t nID, double fMin, double fMax )
{
  tpsVoltagePtr psVoltage = get_voltage( nID );

  if( psVoltage != NULL )
  {
    psVoltage->fRangeMin = fMin;
    psVoltage->fRangeMax = fMax;
  }
  
  return RC_OK;
}

rc_t print_voltage( uint16_t nID, FILE *stream )
{
  rc_t nRC = RC_OK;

  tpsVoltagePtr psVoltage = get_voltage( nID );

  if( psVoltage != NULL )
  {
    if(( stream != NULL ) && (!ferror( stream )))
    {
      fprintf( stream, "<voltage id=%d>", nID );
      fprintf( stream,   "<state>%s</state>", psVoltage->nState == VOLTAGE_ON ? "ON" : "OFF" );
      fprintf( stream,   "<range_min>%lf</range_min>", psVoltage->fRangeMin );
      fprintf( stream,   "<range_max>%lf</range_max>", psVoltage->fRangeMax );
      fprintf( stream, "</voltage>" );
    }
    else
    {
      nRC = RC_ERROR_STREAM_ERROR;
      log_error( nRC, __LINE__ );
    }
  }
  else
  {
    nRC = RC_ERROR_INVALID_IO_ID;
    log_error( nRC, __LINE__ );
  } 

  return nRC;
}

rc_t print_voltages( FILE * stream )
{
  rc_t nRC = RC_OK;

  if(( stream != NULL ) && (!ferror( stream )))
  {
    fprintf( stream, "<voltages amount=%d>", nNrVoltages );

    for( uint16_t nI = 0; nI < nNrVoltages; nI++ )
    {
      nRC = print_voltage( nI, stream );
      if( nRC != RC_OK )
        break;
    }

    fprintf( stream, "</voltages>" );
  }
  else
  {
    nRC = RC_ERROR_STREAM_ERROR;
    log_error( nRC, __LINE__ );
  }

  return nRC; 
}

// --------------------------------------------------------------------------
// FRONTPANEL LAMP CONTROL
// --------------------------------------------------------------------------

tpsLampPtr get_lamp( uint16_t nLampID )
{
  tpsLampPtr psLamp = NULL;

  for( uint16_t nI = 0; nI < nNrLamps; nI++ )
  {
    if( asLamps[ nI ].nID == nLampID )
    {
      psLamp =  &asLamps[ nI ];
      break;
    }
  }

  return psLamp;
}

uint16_t lamp_id_to_i2c_address( uint16_t nLampID )
{
  tpsLampPtr    psLamp    = get_lamp( nLampID );
  tpsI2CCardPtr psI2CCard = NULL;
  uint16_t      nAddress  = 0;

  if( psLamp != NULL )
  {
    psI2CCard = get_i2ccard( psLamp->nI2CCardID );
    
    if( psI2CCard != NULL )
      nAddress = psI2CCard->nOutAddress;
    else
      nAddress = RC_ERROR_INVALID_I2C_ADDRESS;
  }

  return nAddress;
}

rc_t switch_lamp( uint16_t nID, uint8_t nLampStatus )
{
  rc_t nRC = RC_OK;

  if( nLampStatus != LAMP_BLINK )
  {
    nRC = switch_io( lamp_id_to_i2c_address( nID ), nID, nLampStatus );
    if( nRC == RC_OK )
      set_lamp_status( nID, nLampStatus );
  }

  return nRC;
}

rc_t update_lamps( void )
{
  blink_lamps();
  update_lamps_on_frontpanel();

  return RC_OK;
}

rc_t update_lamps_on_frontpanel( void )
{
  rc_t          nRC       = RC_OK;
  tpsLampPtr    psLamp;
  tpsI2CCardPtr psI2CCard;

  for( uint16_t nI = 0; nI < nNrLamps; nI++ )
  {
    psLamp = &asLamps[ nI ];
    psI2CCard = get_i2ccard( psLamp->nI2CCardID );
    if( psI2CCard != NULL )
    {
      nRC= switch_io( psI2CCard->nOutAddress, 
                      psLamp->nID,
                      psLamp->nState );
      if( nRC != RC_OK )
        break;
    } 
  }

  return nRC;
}

rc_t set_lamp_status( uint16_t nID, uint8_t nLampStatus )
{
  rc_t       nRC    = RC_OK;
  tpsLampPtr psLamp;

  if( nID == NO_LAMP )
    return RC_OK;

  psLamp = get_lamp( nID );
  if( psLamp != NULL )
  {
    if( nLampStatus != LAMP_STATE_BLINK )
    {
      psLamp->nState = nLampStatus;
    }
    else
    {
      psLamp->bBlink = TRUE;    
    } 
  }
  else
  {
    nRC = RC_ERROR_INVALID_IO_ID;
    log_error( nRC, __LINE__ );
  }
 
  return nRC;
}

rc_t blink_lamps( void )
{
  rc_t       nRC     = RC_OK;
  tpsLampPtr psLamp;

  static uint32_t nTicks = 0;
  static uint8_t  nState = LAMP_STATE_OFF;

  nTicks++;
 
  if( nTicks == LAMP_BLINK_TICKS )
  {
    if( nState == LAMP_STATE_OFF )
      nState = LAMP_STATE_ON;
    else
      nState = LAMP_STATE_OFF;

#if defined( __DEBUG_BLINK_LAMPS__ )

    if( bit_is_set( PINB, PINB1 ) )
    {
      PORTB = PINB & ~( 1 << PINB1 );
    } 
    else
    {
      PORTB = PINB | (1 << PB1);
    }

#endif

    for( uint16_t nI = 0; nI < nNrLamps; nI++ )
    {
      psLamp = &asLamps[ nI ];
      if( psLamp->bBlink == LAMP_BLINK )
      {
        psLamp->nState = nState;
        
        nRC = switch_io( lamp_id_to_i2c_address( psLamp->nID ),
                         psLamp->nID,
                         nState );
        if( nRC != RC_OK )
          break;
      }
    }
    nTicks = 0;
  }

  return nRC;
}

rc_t print_lamp( uint16_t nID, FILE *stream )
{
  rc_t nRC = RC_OK;

  tpsLampPtr psLamp = get_lamp( nID );

  if( psLamp != NULL )
  {
    if(( stream != NULL ) && (!ferror( stream )))
    {
      fprintf( stream, "<lamp id=%d>", nID );
      fprintf( stream,   "<state>%s</state>", psLamp->nState == LAMP_STATE_ON ? "ON" : "OFF" );
      fprintf( stream,   "<blink>%s</blink>", psLamp->bBlink == LAMP_BLINK ? "YES" : "NO" );
      fprintf( stream, "</lamp>" );
    }
    else
    {
      nRC = RC_ERROR_STREAM_ERROR;
      log_error( nRC, __LINE__ );
    }
  }
  else
  {
    nRC = RC_ERROR_INVALID_IO_ID;
    log_error( nRC, __LINE__ );
  } 
  
  return nRC;
}

rc_t print_lamps( FILE * stream )
{
  rc_t nRC = RC_OK;

  if(( stream != NULL ) && (!ferror( stream )))
  {
    fprintf( stream, "<lamps amount=%d>", nNrLamps );

    for( uint16_t nI = 0; nI < nNrLamps; nI++ )
    {
      nRC = print_lamp( nI, stream );
      if( nRC != RC_OK )
        break;
    }

    fprintf( stderr, "</lamps>" );
  }
  else
  {
    nRC = RC_ERROR_STREAM_ERROR;
    log_error( nRC, __LINE__ );
  }

  return nRC; 
}

// --------------------------------------------------------------------------
// CONTROLLER INITIALIZATION
// --------------------------------------------------------------------------

// --------------------------------------------------------------------------
// INIT_PORT_A                                                               
// --------------------------------------------------------------------------
// Initializes port A as needed                                              
//                                                                           
// Parameters: None                                                          
// --------------------------------------------------------------------------

void init_port_a ( void )
{
  uint8_t nTmpSREG = SREG; // Save status register

  cli();                   // Ensure atomic execution

  ADCSRA  =    0;          // Disable ADC
  DDRA    = 0x00;          // All line input
  PORTA   = 0xFF;          // All internal pullups on

  asm volatile ( "nop" );  // Give the AVR time to settle

  SREG = nTmpSREG;         // We're done so interrupts ok now

  return;
}

// --------------------------------------------------------------------------
// ANALYSE_RESET_REASON                             
// --------------------------------------------------------------------------

void analyse_reset_reason ( void )
{
  char *pcReason;
  
  // Check source of reset

  if (MCUCSR & 1)               // Power-on Reset
  {
    pcReason = acReasonPOR;

    // put POR handler here, if required
  }
  else if (MCUCSR & 2)          // External Reset
  {
    pcReason = acReasonEXTRST;

    // put external reset handler here, if required
  }
  else if (MCUCSR & 4)          // Brown-Out Reset
  {
    pcReason = acReasonBROWNOUT;
  }
  else                          // Watchdog Reset
  {
    pcReason = acReasonWATCHDOG;

    // put watchdog reset handler here, if required
  };

  MCUCSR=0;      // Reset Status Register

  display_format( 1, "RESET REASON:" );  
  display_format( 2, "%s", pcReason );

  return;	
}

// --------------------------------------------------------------------------
// INIT_PORT_B                   
// --------------------------------------------------------------------------
// Initializes port B as needed                                              
//                                                                           
// Parameters: None                                                          
// --------------------------------------------------------------------------

void init_port_b ( void )
{
  uint8_t nTmpSREG = SREG; // Save status register

  cli();                   // Ensure atomic execution

  DDRB |= (1<<PB0);        // Set line 0 of port B as output
  DDRB |= (1<<PB1);        // Set line 1 of port B as output
  DDRB |= (1<<PB2);        // Set line 2 of port B as output
  DDRB |= (1<<PB3);        // Set line 3 of port B as output
  DDRB |= (1<<PB4);        // Set line 4 of port B as output

  // Heartbeat LED: Line 0 -> Switch off initially

  PORTB &= ~( 1<<PB0 );

#if defined( __DEBUG_BLINK_LAMPS__ )

  PORTB &= ~( 1<<PB1 );

#endif

  asm volatile ( "nop" );  // Give the AVR time to settle

  SREG = nTmpSREG;         // We're done so interrupts ok now 

  return;
}

rc_t init_0( void )
{
  rc_t nRC = RC_OK;

  // Init step 0.1: Init ports A and B on Mini-Mega-Board & I2C

  init_port_a();
  init_port_b();

  // Init step 0.2: LCD

  lcd_init( LCD_DISP_ON, I2C_LCD_ADDRESS );

  display_format( 2, "-> PLEASE WAIT !" );

  // Init step 0.3: DARC

  darc_init( DARC_BAUDRATE );

  // Init step 0.4: I2C init and switch all relays off ! 

  if( bUseI2C == TRUE )
  {
    i2c_init();
    _delay_ms( 10 );
    init_relays();
  }

  // Init step 0.5: Prepare error logging to EEPROM

  init_error_logging();

  // And now enable interrupts 

  sei();

  return nRC;
}

rc_t init_1( void )
{
  rc_t     nRC = 0;
  uint8_t  nTmpSREG = SREG;

  cli();    // Ensure atomic execution

  // Clear keys

  clear_key_state();
  
  // Init I/O timer ad start running the ISR

  init_io_timer();

  // ---
  
  SREG = nTmpSREG;

  return nRC;
}

rc_t greet( void )
{
  rc_t nRC = RC_OK;
  
  // Line 1: Title
  //                12345678901234567890
  display_format( 1, "* DARC *" );
  
  // Line 2: Model
  display_format( 2, ">%6s<  by DG1SBG", acPSUModel );
  
  // Line 3 Serial Nr and Version
  display_format( 3, 
                  "# %s - %s", 
                  acPSUSerialNr,
                  acPSUVersion );

  if( use_darc() == TRUE )
  {
    FILE * stream = darc_stream();

    if( (stream != NULL ) && !ferror( stream ) )
    {
      fprintf( stream, "<psu_info>\r\n" );
      fprintf( stream, "  <model>%s</model>\r\n", acPSUModel );
      fprintf( stream, "  <version>%s</version>\r\n", acPSUVersion );
      fprintf( stream, "  <serial_nr%s</serial_nr>\r\n", acPSUSerialNr );
      fprintf( stream, "</psu_info>\r\n" );
    }
    else
    {
      nRC = RC_ERROR_STREAM_ERROR;
    }
  }
  
  if( nRC == RC_OK )
  {
    _delay_ms( 1000 );
  }

  return nRC;
}

// --------------------------------------------------------------------------
// MISC                   
// --------------------------------------------------------------------------

void enable_i2c( void )
{
  bUseI2C = TRUE;
}

void disable_i2c( void )
{
  bUseI2C = FALSE;
}

void set_debug_level( uint16_t nLevel )
{
  nDebugLevel = nLevel;
}

rc_t print_object( uint8_t nObjectClassID, 
                   uint8_t nObjectID,
                   FILE *  stream )
{
  rc_t nRC = RC_OK;

  switch( nObjectClassID )
  {
    case OBJECT_CLASS_ID_RELAY:
      nRC = print_relay( nObjectID, stream );
      break;

    case OBJECT_CLASS_ID_VOLTAGE:
      nRC = print_voltage( nObjectID, stream );
      break;

    case OBJECT_CLASS_ID_FAN:
      nRC = print_fan( nObjectID, stream );
      break;

    case OBJECT_CLASS_ID_LAMP:
      nRC = print_lamp( nObjectID, stream );
      break;

    case OBJECT_CLASS_ID_PSU_CONFIGURATION:
      nRC = print_psu_configuration( stream );
      break;

    case OBJECT_CLASS_ID_FAILURE_REASONCODE:
      nRC = print_failure_reasoncode( stream );
      break;

    case OBJECT_CLASS_ID_STATUS:
      nRC = print_status( stream );
      break;

    case OBJECT_CLASS_ID_ERRORLOG:
      nRC = print_errorlog( stream );
      break;

    default:
      log_error( SW_ERROR, __LINE__ );
      break;
  }

  return nRC;
}

// ==========================================================================
//    M A I N   P S U   F U N C T I O N S
// ==========================================================================

rc_t psu_off( void )
{
  rc_t nRC = RC_OK;

  // Set the status to PSU OFF
  
  clear_display();
  display_format( 1, "Waiting for PWR OFF!" );

  set_status( PSU_STATUS_ID_PSU_OFF );

  // Switch main power off - the Controller will stay on power !

  switch_relay( RELAY_ID_MAIN_POWER, RELAY_OFF );

  // Now wait for a Power On on the main switch

  wait_for_power_off();

  clear_display();

  switch_status( PSU_STATUS_ID_PSU_BOOTING );

  return nRC;
}

rc_t boot( void )
{
  rc_t nRC = RC_OK;

  clear_display();

  display_format( 2, "Waiting for PWR ON !" );

  set_status( PSU_STATUS_ID_PSU_BOOTING );

  // Switch main fan to low speed initially

  nRC = switch_fan( FAN_ID_MAIN_FAN, FAN_SPEED_1 );
  RESET_ON_ERROR(nRC);

  nRC =  bring_down_voltages();
  RESET_ON_ERROR(nRC);

  // Wait for Power On switch being turned on
 
  set_lamp_status( LAMP_ID_MAIN_POWER, LAMP_STATE_BLINK );

  wait_for_power_on();

  display_format( 2, "Powering up ...     " );

  // Finally go to Standby status

  switch_status( PSU_STATUS_ID_PSU_STANDBY );

  return RC_OK;
}

rc_t standby( void )
{
  uint8_t  nTmpSREG = 0;
  uint16_t nPB      = 0;

  clear_display();
  display_format( 2, "Waiting for Mode !" );    

  set_status( PSU_STATUS_ID_PSU_STANDBY );

  while( 1 )
  { 
    BREAK_IF_STATUS_CHANGED();

    // Atomic read of key state

    nTmpSREG = SREG;
    cli();
    
    nPB = ~key_state;
    
    SREG = nTmpSREG;  
     
#if defined( __DEBUG_STANDBY__ )
    display_format( 1, "key_state = %u", key_state );
    display_format( 2, "POWER? %u", POWER_SWITCH_ON(nPB));
    display_format( 3, "RESET? %u", RESET_PB_PRESSED(nPB));
    display_format( 4, "OPER?  %u", OPER_PB_PRESSED(nPB));
    _delay_ms( 50 );
#endif
    
#if !defined( __DEBUG_STANDBY__ )
    
    if( ! POWER_SWITCH_ON(nPB) )
    {
      switch_status( PSU_STATUS_ID_PSU_SHUTTING_DOWN );
      break;   
    }

    if( RESET_PB_PRESSED(nPB) )
    {
      switch_status( PSU_STATUS_ID_PSU_RESETTING );
      break;
    }

    if( OPER_PB_PRESSED(nPB) )
    {
      switch_status( PSU_STATUS_ID_PSU_OPERATING );
      break;    
    }

    if( TEST_PB_PRESSED(nPB) )
    {
      switch_status( PSU_STATUS_ID_PSU_IN_TEST_MODE );
      break;    
    }
#endif
  }  

  return RC_OK;
}

rc_t operate( void )
{
  rc_t     nRC       = RC_OK;
  uint8_t  nTmpSREG  = 0;
  uint16_t nPB       = 0;

  clear_display();

  set_status( PSU_STATUS_ID_PSU_OPERATING );

  // switch on normal operation voltages

  nRC = bring_up_voltages();   

  if( nRC != RC_OK )
  {
    log_error( nRC, __LINE__ );
  }
  else
  {
    display_format( 1, "Running ..." );
     
    while( 1 )
    { 
      BREAK_IF_STATUS_CHANGED();

      // Atomic read of key state

      nTmpSREG = SREG;
      cli();
    
      nPB = ~key_state;
      
      SREG = nTmpSREG;
  
#if defined( __DEBUG_OPERATE__ )

      display_format( 1, "key_state = %u", key_state );
      display_format( 2, "POWER? %u", POWER_SWITCH_ON(nPB));
      display_format( 3, "RESET? %u", RESET_PB_PRESSED(nPB));
      display_format( 4, "OPER?  %u", OPER_PB_PRESSED(nPB));
      _delay_ms( 5 );

#endif

      if( ! POWER_SWITCH_ON(nPB) )
      {
        switch_status( PSU_STATUS_ID_PSU_SHUTTING_DOWN );
        break;   
      } 

      if( RESET_PB_PRESSED(nPB) )
      { 
        nRC = bring_down_voltages();
        switch_status( PSU_STATUS_ID_PSU_RESETTING );
        break;
      }

      if( OPER_PB_PRESSED(nPB) )
      {
        nRC = bring_down_voltages();
        switch_status( PSU_STATUS_ID_PSU_STANDBY );
        break;    
      }

      if( TEST_PB_PRESSED(nPB) )
      {
        nRC = bring_down_voltages();
        switch_status( PSU_STATUS_ID_PSU_IN_TEST_MODE );
        break;    
      }
    }  
  }

  return nRC;
}

rc_t test_mode( void )
{
  rc_t     nRC      = RC_OK;
  uint8_t  nTmpSREG = 0;
  uint16_t nPB      = 0;

  uint8_t       nMode     = VOLTAGE_OFF;
  tpsVoltagePtr psVoltage = NULL;
  
  clear_display();
  display_format( 1, "VOLTAGE STATUS:" );

  set_status( PSU_STATUS_ID_PSU_IN_TEST_MODE );

  nRC = bring_down_voltages();   
  
  if( nRC == RC_OK )
  {
    while( 1 )
    { 
      BREAK_IF_STATUS_CHANGED();

      display_format( 2, 
                      "HVEN= %d Ua = %d",
                      (get_relay( RELAY_ID_HV_ENABLE ))->nState,
                      (get_voltage( VOLTAGE_ID_UA ))->nState );
      display_format( 3, 
                      "Ug2 = %d Ug1= %d Uh= %d",
                      (get_voltage( VOLTAGE_ID_UG2 ))->nState,
                      (get_voltage( VOLTAGE_ID_UG1 ))->nState,
                      (get_voltage( VOLTAGE_ID_UH ))->nState );

      // Atomic read of key state

      nTmpSREG = SREG;
      cli();
    
      nPB = ~key_state;
      
      SREG = nTmpSREG;
  
      if( ! POWER_SWITCH_ON(nPB) )
      {
        switch_status( PSU_STATUS_ID_PSU_SHUTTING_DOWN );
        break;   
      } 

      if( RESET_PB_PRESSED(nPB) )
      { 
        switch_status( PSU_STATUS_ID_PSU_RESETTING );
        break;
      }

      if( OPER_PB_PRESSED(nPB) )
      {
        bring_down_voltages();
        switch_status( PSU_STATUS_ID_PSU_OPERATING );
        break;    
      }

      if( TEST_PB_PRESSED(nPB) )
      {
        bring_down_voltages();
        switch_status( PSU_STATUS_ID_PSU_STANDBY );
        break;    
      }

      if( HV_ENABLE_PB_PRESSED(nPB) )
      {
        toggle_relay( RELAY_ID_HV_ENABLE );        
      }

      if( UA_PB_PRESSED(nPB) )
      {
        psVoltage = get_voltage( VOLTAGE_ID_UA );
        if( psVoltage != NULL )
        {
          if( psVoltage->nState == VOLTAGE_ON )
            nMode = VOLTAGE_OFF;
          else
            nMode = VOLTAGE_ON;

          toggle_voltage( VOLTAGE_ID_UA );
        }
      }

      if( UG2_PB_PRESSED(nPB) )
      {
        psVoltage = get_voltage( VOLTAGE_ID_UG2 );
        if( psVoltage != NULL )
        {
          if( psVoltage->nState == VOLTAGE_ON )
            nMode = VOLTAGE_OFF;
          else
            nMode = VOLTAGE_ON;

          toggle_voltage( VOLTAGE_ID_UG2 );
        }
      }

      if( UG1_PB_PRESSED(nPB) )
      {
        psVoltage = get_voltage( VOLTAGE_ID_UG1 );
        if( psVoltage != NULL )
        {
          if( psVoltage->nState == VOLTAGE_ON )
            nMode = VOLTAGE_OFF;
          else
            nMode = VOLTAGE_ON;

          toggle_voltage( VOLTAGE_ID_UG1 );
        }
      }

      if( UH_PB_PRESSED(nPB) )
      {
        psVoltage = get_voltage( VOLTAGE_ID_UH );
        if( psVoltage != NULL )
        {
          if( psVoltage->nState == VOLTAGE_ON )
            nMode = VOLTAGE_OFF;
          else
            nMode = VOLTAGE_ON;

          toggle_voltage( VOLTAGE_ID_UH );
        }
      }
    }  
  }

  bring_down_voltages();

  return nRC;
}

rc_t fail( uint16_t nReasonCode )
{
  rc_t nRC = RC_OK;

  uint8_t  nTmpSREG;
  uint16_t nPB;
 
  uint16_t nCountdown = PSU_FAIL_COOLDOWN_TICKS;
  
  // First shutdown all voltages !!!
 
  bring_down_voltages();
  
  // Switch fan to high speed to really cool the PSU down

  switch_fan( FAN_ID_MAIN_FAN, FAN_SPEED_MAX );

  // Due to the requirement of first shutting down the voltages
  // we set the status display only now ...

  clear_display();
  set_status( PSU_STATUS_ID_PSU_FAILURE );

  // Set FAIL lamp to blink

  set_lamp_status( LAMP_ID_FAIL, LAMP_BLINK );

  // Log Reason Code as error
  
  log_error( nReasonCode, __LINE__ );
  
  // ... and display Reason Code 

  display_format( 1, "FAILURE ! ->" );
  display_format( 2, "Reason Code: %04d", get_failure_reasoncode() );

  darc_send_failure_reasoncode();

  // In the FAIL status we only accept a reset or a power down!

  while( 1 )
  { 
    // Countdown Fan being on full tilt ...

    if( nCountdown <= 0 )
      switch_fan( FAN_ID_MAIN_FAN, FAN_SPEED_1 );
    else
      nCountdown--;
  
    // Atomic read of key state (16 bit value)

    nTmpSREG = SREG;
    cli();
    
    nPB = ~key_state;
    
    SREG = nTmpSREG;  
     
#if defined( __DEBUG_FAIL__ )
    display_format( 1, "key_state = %u", key_state );
    display_format( 2, "POWER? %u", POWER_SWITCH_ON(nPB));
    display_format( 3, "RESET? %u", RESET_PB_PRESSED(nPB));
    _delay_ms( 50 );
#endif
  
#if !defined( __DEBUG_FAIL__ )
    if( ! POWER_SWITCH_ON(nPB) )
    {
      switch_status( PSU_STATUS_ID_PSU_SHUTTING_DOWN );
      break;   
    }

    if( RESET_PB_PRESSED(nPB) )
    {
      switch_status( PSU_STATUS_ID_PSU_RESETTING );
      break;
    }
#endif
  }  

  return nRC;
}

rc_t shutdown( void )
{
  rc_t    nRC        = RC_OK;

#if (PSU_SHUTDOWN_COOLDOWN_TIME > 60 ) || (PSU_SHUTDOWN_COOLDOWN_TIME < 10)
#error "->> Shutdown time should be within 10s and 60s !"
#endif

  uint8_t nCountdown = PSU_SHUTDOWN_COOLDOWN_TIME;

  clear_display();
  display_format( 1, "Shutting down ..." );

  // set_status( STATUS_ID_PSU_SHUTTING_DOWN );

  // switch off all voltages

  switch_off_all_voltages();

  switch_fan( FAN_ID_MAIN_FAN, FAN_SPEED_MAX );

  while( nCountdown > 0 )
  {
    //                  12345678901234567890
    display_format( 2, "Still %3u s until", nCountdown );
    display_format( 3, "PSU is switched off!" );

    nCountdown--; 
    _delay_ms( 4 ); // Really does wait much longer ...
  }
  
  // switch off fan and go to PSU OFF mode

  switch_fan( FAN_ID_MAIN_FAN, FAN_OFF );

  switch_status( PSU_STATUS_ID_PSU_OFF );

  return nRC;
}

// --------------------------------------------------------------------------
// I S R s & Utils
// --------------------------------------------------------------------------

ISR ( TIMER1_COMPA_vect  )  // The main routine triggered by Timer 1 to
{                           // interact with the user
  int nRC = 0;

  // uint8_t nTmpSREG = SREG;
  // cli();

  nRC = lock();

  if( nRC == RC_OK )
  {
    heart_beat();

    // sample_voltages();

    debounce(); // Debounce pushbuttons

    process_darc();

    update_lamps();
     
    unlock();
  }

  // SREG = nTmpSREG;

  return;
}

ISR ( INT0_vect )   // FAILURE INTERRUPT -> Switch to FAILURE status
{
  if( bInStatusNetEngine == TRUE )
    switch_status( PSU_STATUS_ID_PSU_FAILURE );
  else
    nCmdFromISR = ISR_DO_FAIL;
}

ISR ( INT1_vect )   // SHUTDOWN INTERRUPT -> Switch to SHUTDOWN status
{
  if( bInStatusNetEngine == TRUE )
    switch_status( PSU_STATUS_ID_PSU_SHUTTING_DOWN );
  else
    nCmdFromISR = ISR_DO_SHUTDOWN;
}

ISR ( __vector_default )
{
  // Reaching this point indiciates we have a bug/mismatch between H/W and
  // S/W - we therefore reboot ...

  log_error( SW_ERROR, __LINE__ );
  reset();

	// This point is never reached
}

void execute_cmd_from_isr( void )
{
  switch( nCmdFromISR )
  {
    case ISR_DO_FAIL:
      nCmdFromISR = NO_CMD;
      nFailureReasonCode = PSU_FAILURE_REASON_EXTERNAL_FAILURE;
      fail( PSU_FAILURE_REASON_EXTERNAL_FAILURE );
      break;

    case ISR_DO_SHUTDOWN:
      nCmdFromISR = NO_CMD;
      shutdown();
      break;

    case ISR_DO_RESET:
      nCmdFromISR = NO_CMD;
      reset();
      break;

    default:
      // Do nothing
      break;
  }

  nCmdFromISR = NO_CMD;
}

// --------------------------------------------------------------------------
// M A I N
// --------------------------------------------------------------------------

int main ( void )
{
  rc_t nRC = RC_OK;

  _delay_ms( 2000 );

  // "Configure" communication

  enable_i2c();
  enable_darc();

  nNrStatuses            =  8;
  nNrI2CCards            =  1;
  nNrRelays              =  7;
  nNrFans                =  1;
  nNrVoltages            =  4;
  nNrLamps               = 10;
  nNrPBs                 =  9;
  nNrVoltageCombinations =  0;

  // Status engine: Still outside of it

  bInStatusNetEngine = FALSE;

  // Status engine and Debug Level init
    
  set_debug_level( PSU_DEBUG_LEVEL_VERBOSE_2 );

  // Do basic inits as fast as possible

  init_0();

  // execute_cmd_from_isr();
  
  // Display the reason why the controller is booting

  analyse_reset_reason();

  // execute_cmd_from_isr();

  _delay_ms( 1000 );

  // Say hello on LCD

  greet();
  // execute_cmd_from_isr();

  _delay_ms( 1000 );

  // Make error logging in EEPROM possible
  
  show_last_logged_error(); 
  // execute_cmd_from_isr();

  _delay_ms( 1000 );
   
  // Kick off status net - THIS IS THE MAIN LOOP ...
  
  bInStatusNetEngine = TRUE;
  
  // Next level inits

  init_1();

  clear_display();

  set_status( PSU_STATUS_ID_PSU_OFF );

  nRC = switch_status( PSU_STATUS_ID_PSU_BOOTING );
  
  while( nRC == RC_OK )
  { 
    nRC = goto_next_status( nNextStatus );
  }

  // When this point is reached it most likely means that a reset is 
  // required !

  if( IS_ERROR( nRC ) )
    log_error( nRC, __LINE__ );

  reset(); // Reboot PSU Controller !!!

  for( ;; ){} // forever

  // This point never reached ...
  return 1; 
}
